private import sdl;
version(use_mixer)  private import sdl_mixer;
version(use_opengl) { 
    private import opengl;
    private import openglu;
    private import util_opengl.expansion;
    private import util_opengl.matrixsaver;
    private import util_opengl.texture;
    private import util_opengl.light;
    private import util_opengl.material;
    private import util_opengl.array;
    private import util_opengl.model;
}
private import util_sdl.mainfiber;
private import util_sdl.fiber;
private import util_sdl.controler;


private import std.stdio;
private import std.string;
private import std.math;
private import util.math;
private import util.vector;
private import util.color;
private import util.rand;


// constants
const int   BULLET_SMOKE_LENGTH  = 60;
const int   MISSILE_SMOKE_LENGTH = 40;
const int   FAR_LIMIT = 30000;
const int   LOCK_MAX  = 8;
const float LOCKING_Z_LIMIT = -500;




// globals
CCharTable  font;
CLight      light;
Rand        rand;
CGradation  gradSmoke;
CGradation  gradDestroy;

CFlyer      flyer;
CSite       site;
CCamera     camera;
CSky        sky;
IActor      draw_chain_farther;
IActor      draw_chain_nearest;




// entry point
int main(char[][] arg)
{
    (new Test).execute(arg);
    return 0;
}




// Main fiber
class Test : IMainFiber
{
public:
    void enterInstance(char[][] arg)
    {
        setWindow("ABCD_test", 320, 240, 60, 5);
        parseArg(arg);
        moduleControl.setSystemHalt(SDLK_ESCAPE);
        moduleControl.getDefaultControler().assign(SDLK_F1, PAD.BUTTON10);
        moduleFiber.add(new ScoreManager);
        moduleFiber.add(new ObjectManager);
        moduleFiber.add(new ScreenManager);
    }
    
    void enter()
    {
        glClearColor(0.5882, 0.6078, 0.6824,1);
        setFog(0, 1500, 0xc0aca9);

        rand = new Rand();
        font = new CCharTable((new CTexture(0xffffff, "resource\\font.bmp")).setFormat(false,true,false,false,false).register(), 8, 8, 0);
        light = new CLight(0, -1, -1);
    }
    
    void run()
    {
        if (moduleControl.isHitted(PAD.BUTTON10)) {
        }
    }
    
    void draw()
    {
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    }
}




// Score manager
class ScoreManager : IFiber
{
public:
    static uint score;
    static uint miss_count;
    
    static void add(int bonus)
    {
        score += bonus;
    }

    static void miss()
    {
        miss_count++;
    }
    
    void enterScene()
    {
        score = 0;
        miss_count = 0;
    }
}



// Screen manager
class ScreenManager : IFiber
{
public:
    static char[32] buf;
    static char[]   msg;
    static float    alpha;
    
    void enter()
    {
        camera.pos.z = setScreenPers(60, 640, 480, -480, FAR_LIMIT);
        msg.length = 0;
        alpha = 0;
    }
    
    static void print(char[] str)
    {
        msg = buf[0..str.length];
        msg[] = str;
        alpha = 1;
    }
    
    void run()
    {
        alpha-=0.01;
        if (alpha < 0) msg.length = 0;
    }
    
    void draw()
    {
        char[16] buf;
        char[]   str;
        setSpriteAttribution();
        setAlphaBlend();
        pushScreen(320, 240);
        if (msg.length != 0) {
            glColor4f(1,1,1,alpha);
            font.draw(-48, -60, msg);
        }
        glColor4f(1,1,1,1);
        str = sformat(buf, "FPS:%.1f", moduleFPS.getFPS());
        font.draw(96, -112, str);
        str = sformat(buf, "LOCKED:%02d", site.lock_count);
        font.draw(-159, -112, str);
        glColor4f(0,0,0,1);
        str = sformat(buf, "SCORE:%08d", ScoreManager.score);
        font.draw(-159, 119, str);
        str = sformat(buf, "MISS :%d",   ScoreManager.miss_count);
        font.draw(-159, 111, str);
        popScreen();
    }
}




// Object manager
class ObjectManager : IFiber
{
public:
    int         rolling_count;
    int         rolling_input;

public:
    void enter()
    {
        CTexture    tex    = new CTexture("resource/particle_texture.bmp", 3);
        CSolidModel mdlFly = new CSolidModel("resource/flyer.mqo");
        CSolidModel mdlMsl = new CSolidModel("resource/missile.mqo");
        CSolidModel mdlEne = new CSolidModel("resource/enemy.mqo");
        CSolidModel mdlBul = new CSolidModel(mdlMsl);
        CTexture    texSmk = new CTexture("resource/smoke.bmp", "resource/smoke.bmp");
        CSprite     sprSmk = new CSprite(texSmk.setFormat(true,true,false,false,false));
        CSprite     sprTex = new CSprite(tex);
        CTexture    texExp = new CTexture("resource/explosion.bmp", "resource/_explosion.bmp");//3);
        CSprite     sprExp = new CSprite(texExp.setFormat(true,true,true,false,false));
        
        mdlBul.rotateY(180);
        sprSmk.setUV(0, 0, 1, 1);
        
        CFlyer.mdl   = mdlFly.compile();
        CFlyer.spr   = sprTex;
        CEnemy.mdl   = mdlEne.compile();
        CEnemy.spr   = sprTex;
        CMissile.mdl = mdlMsl.compile();
        CBullet.mdl  = mdlBul.compile();
        CSmoke.spr   = sprSmk;
        CSite.spr    = sprTex;
        CExplosion.spr = sprExp;
        
        camera = new CCamera();
        flyer  = new CFlyer();
        site   = new CSite();
        sky    = new CSky();
        CMissile(32);
        CBullet(64);
        CEnemy(64);
        CSmoke(2048);
        CExplosion(32);
        CExplosionParticle(128);

        gradSmoke   = (new CGradation(0xffffffff,0x00ffffff)).add(0.5,0xffffffff).update();
        gradDestroy = (new CGradation(0x00000000,0x00202f30)).add(0.15,0xff404040).update();
        
        draw_chain_farther = new IActor;
        draw_chain_nearest = new IActor;
        draw_chain_farther.pos.z = -float.infinity;
        draw_chain_nearest.pos.z = float.infinity;
        draw_chain_farther.drawChainNext = draw_chain_nearest;
        draw_chain_farther.drawChainPrev = draw_chain_nearest;
        draw_chain_nearest.drawChainNext = draw_chain_farther;
        draw_chain_nearest.drawChainPrev = draw_chain_farther;
    }

    void enterScene()
    {
        flyer.initialize();
        flyer.insert_before(draw_chain_nearest);
        rolling_count = 0;
        rolling_input = 0;
    }
    
    void run()
    {
        static int prev_x;
        int x, y;
        
        // counter decrement
        if (rolling_count>0) rolling_count--;
        if (rolling_input>0) rolling_input--;
        
        // controling
        if (!flyer.isDestroyed()) {
            moduleControl.getXY(x, y);
        } else {
            y=1;
        }
        if (rolling_count==0 && x!=0 && prev_x!=x) {
            rolling_input += 10;
            prev_x = x;
            if (rolling_input>20) {
                camera.rolling(prev_x);
                flyer.rolling();
                rolling_count = 80;
                rolling_input = 0;
            }
        }

        // enemy appearance
        static int cnt_enemy = 0;
        if (++cnt_enemy > 20) {
            if (rand.geti(2)==0) {
                CEnemy().initialize(rand.getsf(3000), rand.getsf(1000), -FAR_LIMIT, 50, 60).insert_after(draw_chain_farther);
            }
            cnt_enemy = 0;
        }
        
        // Fire
        if (moduleControl.isHitted(PAD.BUTTON0) && !site.isFiring()) {
            site.fire();
        }
        
        // run object
        camera.run(x, y);
        flyer.run(x, y, moduleControl.isPressed(PAD.BUTTON1), camera.posture.z, (rolling_count>50)?0:(50-rolling_count)*0.02);
        site.run();
        CEnemy.runAll();
        CBullet.runAll();
        CMissile.runAll();
        CExplosion.runAll();
        CExplosionParticle.runAll();
        CSmoke.runAll();
        sky.run();
        
        // locking check
        foreach(enemy; CEnemy) {
            if (enemy.testLock()) {
                site.lock(enemy);
            }
        }
    }
    
    void draw()
    {
        // draw object
        camera.push();
        sky.draw();
        glClear(GL_DEPTH_BUFFER_BIT);
        for (IActor act=draw_chain_farther.drawChainNext; act!=draw_chain_nearest; act=act.drawChainNext) { act.draw(); }
        camera.pop();
        
        pushScreen(640, 480);
        site.draw();
        popScreen();
    }
}




// Actor
class IActor
{
public:
    vector3 pos;
    vector3 vel;
    vector3 acc;
    vector3 posture;
    IActor  drawChainNext;
    IActor  drawChainPrev;

    // drawing task chain operation
    void insert_before(IActor act)
    {
        drawChainPrev = act.drawChainPrev;
        drawChainNext = act;
        act.drawChainPrev.drawChainNext = this;
        act.drawChainPrev = this;
    }

    void insert_after(IActor act)
    {
        drawChainNext = act.drawChainNext;
        drawChainPrev = act;
        act.drawChainNext.drawChainPrev = this;
        act.drawChainNext = this;
    }
    
    void remove()
    {
        if (drawChainNext !is null) drawChainNext.drawChainPrev = drawChainPrev;
        if (drawChainPrev !is null) drawChainPrev.drawChainNext = drawChainNext;
    }

    void sortZ()
    {
        if (drawChainPrev.pos.z > pos.z) {
            remove();
            IActor tgt = drawChainPrev.drawChainPrev;
            while (tgt.pos.z > pos.z) tgt = tgt.drawChainPrev;
            insert_after(tgt);
        } else
        if (drawChainNext.pos.z < pos.z) {
            remove();
            IActor tgt = drawChainNext.drawChainNext;
            while (tgt.pos.z < pos.z) tgt = tgt.drawChainNext;
            insert_before(tgt);
        }
    }

    void draw() { }
};




// actor template
template TInstanceFunction(T)
{
    static T[] instance;
    static int start_index;
    static T   dummy;
    
    bool       isActive = false;
    
    static void opCall(int count)
    {
        dummy = new T();
        instance = new T[count];
        foreach(inout msl; instance) { msl = new T(); }
    }
        
    static T opCall()
    {
        int filter = instance.length-1;
        int idx;
        for(int i=0; i<instance.length; i++) {
            idx = (i+start_index)&filter;
            if (!instance[idx].isActive) {
                instance[idx].isActive = true;
                start_index = (idx+1)&filter;
                return instance[idx];
            }
        }
        return dummy;
    }
    
    static int opApply(int delegate(inout T) dg)
    {
        int result = 0;
        for(int i=0; i<instance.length && result==0; i++) {
            if (instance[i].isActive) result = dg(instance[i]);
        }
        return result;
    }
    
    // operation
    void free()
    {
        isActive = false;
        remove();
    }
    
    static void runAll()
    {
        dummy.free();
        foreach(inst; instance) { if (inst.isActive) inst.run(); }
    }

    static void drawAll()
    {
        dummy.free();
        foreach(inst; instance) { if (inst.isActive) inst.draw(); }
    }
};




// camera
class CCamera : public IActor
{
public:
    matrix4  mat;
    matrix4  mat_inv;
    this() { } 
    
    void rolling(int cwrot)
    {
        posture.z -= cwrot*fRD;
    }
    
    void world2screen(vector2* screen, float x, float y, float z)
    {
        vector3 vc;
        set(&vc, x, y, z);
        mul(&vc, &mat_inv);
        float pers = pos.z/(vc.z+pos.z);
        screen.x = vc.x*pers;
        screen.y = vc.y*pers;
    }

    void run(int x, int y)
    {
        // posture(pitch, head, bank)
        add  (&posture, degree2radian(y), degree2radian(x)*2, degree2radian(x));
        scale(&posture, 0.9, 0.9, 0.95);
        
        loadIdentity(&mat);
        rotateY(&mat, posture.y);
        rotateX(&mat, posture.x);
        rotateZ(&mat, posture.z);

        loadIdentity(&mat_inv);
        rotateX(&mat_inv, -posture.x);
        rotateY(&mat_inv, -posture.y);
        rotateZ(&mat_inv,  posture.z);

    }

    void push()
    {
        glMatrixMode(GL_PROJECTION);
        glPushMatrix();
        glMultMatrixf(&mat.a00);
        glMatrixMode(GL_MODELVIEW);
        glPushMatrix();
    }
    
    void pop()
    {
        glPopMatrix();
        glMatrixMode(GL_PROJECTION);
        glPopMatrix();
        glMatrixMode(GL_MODELVIEW);
    }
}




// flyer
class CFlyer : public IActor
{
public:
    static CSolidModel mdl;
    static CSprite     spr;
    vector3 global_vel;

    enum STATUS {
        TRANSPARENT,
        ALIVE,
        ROLLING,
        DESTROYED,
        REBIRTH,
        max
    }
    STATUS              status;
    int                 status_counter;
    STATUS[STATUS.max]  status_next         = [STATUS.ALIVE, STATUS.ALIVE, STATUS.ALIVE, STATUS.REBIRTH, STATUS.TRANSPARENT];
    int[STATUS.max]     status_counter_next = [120,          0,            60,           120,            120];
    

    void initialize()
    {
        set(&pos, 0, 0, 0);
        set(&vel, 0, 0, 0);
        set(&acc, 0, 0, 0);
        site.initialize();
        status = STATUS.TRANSPARENT;
        status_counter = 120;
    }
    
    
    void run(int x, int y, int accel, float rol, float dec)
    {
        // status
        --status_counter;
        if (status_counter == 0) {
            change_status(status_next[status]);
        }
        
        if (isDestroyed()) {
            // physics
            add(&pos, 0, -5, -20);
            // posture(bank, head, pitch)
            set(&posture, posture.x+10, 0, -36+status_counter*0.3);
            sortZ();

            // world velocity
            global_vel.x *= 0.98;
            global_vel.z *= 0.98;

            // falling
            if (status==STATUS.DESTROYED) {
                CSmoke().initialize(pos.x, pos.y, pos.z, 60, 60, gradDestroy).insert_after(this);
                if (status_counter==1) {
                    CExplosionParticle().initialize(pos.x, pos.y, pos.z, 500).insert_before(this);
                    CExplosion().initialize(pos.x, pos.y, pos.z, 1600).insert_after(this);
                }
            }
        } else {
            // control
            set(&acc.v2, x*3*dec, -y*2*dec);
            vel.z = (accel)?60:45;

            // physics
            add(add(&pos.v2, &vel.v2), &acc.v2, 0.5);
            add(&vel.v2, &acc.v2);
            add(scale(&vel.v2, 0.85), &pos.v2, -0.01);
            // posture(bank, head, pitch)
            set(&posture, -vel.x*4-pos.x*0.1-rol, -vel.x*2-pos.x*0.1, vel.y*2);
            // world velocity
            set(&global_vel, -pos.x*0.1, 0, vel.z);
        }
            
        // site position
        camera.world2screen(&site.window_pos, pos.x, pos.y, 400);
    }
    
    void draw()
    {
        if (status == STATUS.REBIRTH || (status == STATUS.TRANSPARENT && (status_counter & 4)==0)) return;
        
        glDisable(GL_FOG);
        glPushMatrix();
            setModelAttribution();
            glTranslatef(pos.x, pos.y, pos.z);
            glRotatef(posture.z, 1, 0, 0);  // pitch
            glRotatef(posture.y, 0, 1, 0);  // head
            glRotatef(posture.x, 0, 0, 1);  // bank
            mdl.callList();
        
            setSpriteAttribution();
            setAddBlend();
            spr.setUV(0.5,0.5,0.125,0.125);
            spr.set(16, 16);
            spr.translate(24, 4.5, 96);
            for (int i=16; i>0; i--) {
                glColor4f(0.3, 0.5, 1, i*0.015);
                spr.draw();
                spr.translate(-48,0,0);
                spr.draw();
                spr.translate( 48,0,4);
            }
        glPopMatrix();
    }
    
    void change_status(STATUS sts)
    {
        if (status == STATUS.REBIRTH) { initialize(); }
        status = sts;
        status_counter = status_counter_next[sts];
    }

    void rolling()
    {
        change_status(STATUS.ROLLING);
    }
    
    // call from CBullet.run()
    void destroyed()
    {
        change_status(STATUS.DESTROYED);

        CExplosionParticle().initialize(pos.x, pos.y, -500, 60).insert_before(this);
        CExplosion().initialize(pos.x, pos.y, -500, 200).insert_before(this);
        ScoreManager.miss();
    }
    
    bool isAlive()
    {
        return (status == STATUS.ALIVE);
    }

    bool isDestroyed()
    {
        return (status == STATUS.DESTROYED || status == STATUS.REBIRTH);
    }
}




// site
class CSite
{
public:
    static CSprite spr;

    struct LOCKON_SITE {
        CEnemy  enemy = null;
        int     timer;
    }

    vector2                 window_pos;
    vector2                 site_pos;
    LOCKON_SITE[LOCK_MAX]   lockon;
    int                     lock_count;
    int                     firing_counter;
    
    void initialize()
    {
        set(&window_pos, 0, 0);
        set(&site_pos,   0, 0);
        lock_count = 0;
        firing_counter = 0;
        for(int i=0; i<LOCK_MAX; i++) {
            lockon[i].enemy = null;
            lockon[i].timer = 0;
        }
    }
    
    void run()
    {
        // control site
        vector2 vc;
        float   len;
        len = length(sub(&vc, &site_pos, &window_pos));
        if (len > 5) sub(&site_pos, scale(&vc, 5/len));
        else         sub(&site_pos, &vc);
        
        // fire
        if (firing_counter > 0) {
            if (--firing_counter == 0) {
                if (create_missile()) {
                    firing_counter = 6;
                }
            }
        }
        
        // check enemy activity
        lock_count = 0;
        for(int i=0; i<LOCK_MAX; i++) {
            if (lockon[i].enemy !is null) {
                if (lockon[i].enemy.isActive && lockon[i].enemy.pos.z < LOCKING_Z_LIMIT) {
                    ++lock_count;
                } else {
                    lockon[i].enemy = null;
                }
            }
        }
    }
    
    void draw()
    {
        if (flyer.isDestroyed()) return;
        
        setSpriteAttribution();
        setAddBlend();
        glColor4f(1,1,1,1);
        
        spr.set(64, 64);
        spr.translate(window_pos.x, window_pos.y);
        spr.setUV(0, 0.625, 0.125, 0.125);
        spr.draw();

        spr.set(64, 64);
        spr.translate(site_pos.x, site_pos.y);
        spr.setUV(0, 0.5, 0.125, 0.125);
        spr.draw();
        
        spr.setUV(0.125, 0.625, 0.125, 0.125);
        for(int i=0; i<LOCK_MAX; i++) {
            if (lockon[i].enemy !is null) {
                ++lockon[i].timer;
                spr.set(64, 64);
                if (lockon[i].timer < 10) {
                    float scl = (15.0-lockon[i].timer)/5.0;
                    spr.scale(scl, scl);
                    spr.rotate((10-lockon[i].timer)*fPI*0.05);
                    glColor4f(1,1,1,lockon[i].timer*0.1);
                } else {
                    glColor4f(1,1,1,0.25);
                }
                spr.translate(lockon[i].enemy.onScreen.x, lockon[i].enemy.onScreen.y);
                spr.draw();
            }
        }
    }
    
    // call from Test.run()
    bool lock(CEnemy enemy)
    {
        for(int i=0; i<LOCK_MAX; i++) {
            if (lockon[i].enemy is null) {
                lockon[i].enemy = enemy;
                lockon[i].enemy.locked();
                lockon[i].timer = 0;
                return true;
            }
        }
        return false;
    }
    
    // call from Test.run()
    bool fire()
    {
        if (lock_count == 0) return false;
        firing_counter = 1;
        int ratio = 1;
        for(int i=0; i<LOCK_MAX; i++) {
            if (lockon[i].enemy !is null) {
                lockon[i].enemy.fired(ratio);
                ratio *= 2;
            }
        }
        return true;
    }
    
    bool isFiring()
    {
        return (firing_counter>0);
    }
    
    // call from CSite.run()
    bool create_missile()
    {
        for(int i=0; i<LOCK_MAX; i++) {
            if (lockon[i].enemy !is null) {
                float cannon = (lock_count & 1) ? -100 : 100;
                CMissile().initialize(flyer.pos.x+cannon, flyer.pos.y, 0, lockon[i].enemy)
                          .insert_before(draw_chain_nearest);
                lockon[i].enemy = null;
                return true;
            }
        }
        return false;
    }
}



// my missile
class CMissile : public IActor
{
public:
    static CSolidModel mdl;
    
    mixin TInstanceFunction!(CMissile);

    CEnemy target = null;
    int    timer;
    
    CMissile initialize(float x, float y, float z, CEnemy enemy)
    {
        set(&pos, x, y, z);
        set(&vel, 0, 0, -60);
        set(&acc, 0, 0, -2);
        target = enemy;
        timer = 0;
        return this;
    }
    
    void run()
    {
        // control
        if (++timer==60) acc.z=0;

        if (target is null) {
            set(&vel.v2, 0, 0);
            acc.y=5;
        } else {
            vector3 require_vel;
            if (pos.z <= target.pos.z) {
                // achieve to target
                if (target.testHit(&pos.v2)) {
                    // hit !!
                    target.destroyed();
                    free();
                    return;
                }
                target = null;
                copy(&require_vel, &vel);
            } else {
                // frame count to achieve player z position.
                float frm_inv = (target.vel.z + flyer.global_vel.z - vel.z)/(target.pos.z-pos.z);
                // require_vel; required velocity to aim player
                set(&require_vel, (pos.x-target.pos.x)*frm_inv-flyer.global_vel.x, (pos.y-target.pos.y)*frm_inv, 0);
            }
            sub(&acc.v2, &require_vel.v2, &vel.v2);
        }
        
        // physics
        add(&pos, add(&vel, &acc));
        add(&pos, &flyer.global_vel);
        sortZ();
        
        // posture(bank, head, pitch)
        vector3 dir;
        normalize(set(&dir, vel.x, vel.y, vel.z-flyer.vel.z));
        set(&posture, 0, radian2degree(asin(dir.x)), radian2degree(asin(dir.y)));

        // smoke
        float dst = 160;
        CSmoke().initialize(pos.x-dir.x*dst, pos.y-dir.y*dst, pos.z-dir.z*dst, MISSILE_SMOKE_LENGTH, 160, gradSmoke)
                .insert_after(this);

        // disapeared
        if (pos.z < -FAR_LIMIT || pos.y > 10000) free();
    }
    
    void draw()
    {
        setModelAttribution();
        glDisable(GL_FOG);
        glPushMatrix();
            glTranslatef(pos.x, pos.y, pos.z);
            glRotatef(posture.z, 1, 0, 0);  // pitch
            glRotatef(posture.y, 0, 1, 0);  // head
            mdl.callList();
        glPopMatrix();
    }
}




// enemy
class CEnemy : public IActor
{
public:
    static CSolidModel mdl;
    static CSprite     spr;
    
    mixin TInstanceFunction!(CEnemy);
    
    enum STATUS {
        TRANSPARENT,
        ALIVE,
        ESCAPING,
        LOCKED,
        FIRED
    }
    int     timer;
    int     fire_time;
    STATUS  status;
    vector2 onScreen;
    int     bonus;
    
    CEnemy initialize(float x, float y, float z, float v, int fired)
    {
        set(&pos, x, y, z);
        set(&vel, 0, 0, v);
        timer = -30;
        fire_time = fired;
        status = STATUS.TRANSPARENT;
        bonus = 100;
        return this;
    }

    void run()
    {
        // control
        ++timer;
        if (timer == 0) { 
            status = STATUS.ALIVE;
        } else 
        if (timer == fire_time && status != STATUS.TRANSPARENT) {
            CBullet().initialize(pos.x, pos.y, pos.z, 100, 2).insert_after(this);
        }
        
        // physics
        add(add(&pos, &vel), &acc, 0.5);
        add(&vel, &acc);
        add(&pos, &flyer.global_vel);
        sortZ();

        // disapeared
        if (pos.z > 500) free();
        
        // screen coordination
        camera.world2screen(&onScreen, pos.x, pos.y, -pos.z);
    }
    
    void draw()
    {
        setModelAttribution();
        glDisable(GL_FOG);
        glPushMatrix();
            glTranslatef(pos.x, pos.y, pos.z);
            mdl.callList();
        glPopMatrix();

        switch (status) {
        case STATUS.ALIVE:
            setSpriteAttribution();
            setAddBlend();
            pushScreen(640,480);
            spr.set(onScreen.x-32, onScreen.y-32, 64, 64);
            spr.setUV(0.125, 0.5, 0.125, 0.125);
            spr.draw();
            popScreen();
            break;
        case STATUS.FIRED:
            setSpriteAttribution();
            setAddBlend();
            pushScreen(640,480);
            spr.set(onScreen.x-16, onScreen.y-16, 32, 32);
            spr.setUV(0.125, 0.625, 0.125, 0.125);
            spr.draw();
            popScreen();
            break;
        default:
            break;
        }
    }
    
    // call from Test.run()
    bool testLock()
    {
        return ((status == STATUS.ALIVE || status == STATUS.ESCAPING) &&
                 calcDistance2(&site.site_pos, &onScreen) < 256);
    }
    
    // call from CMissile.run()
    bool testHit(vector2* center)
    {
        return (calcDistance2(center, &pos.v2) < 10000);
    }
    
    // call from CSite.lock()
    void locked()
    {
        status = STATUS.LOCKED;
    }
    
    // call from CSite.fire()
    void fired(int ratio)
    {
        status = STATUS.FIRED;
        bonus *= ratio;
    }
    
    // call from CMissile.run()
    void destroyed()
    {
        status = STATUS.TRANSPARENT;
        CExplosionParticle().initialize(pos.x, pos.y, pos.z, 500).insert_before(this);
        CExplosion().initialize(pos.x, pos.y, pos.z, 1600).insert_before(this);
        ScoreManager.add(bonus);
        ScreenManager.print("BONUS : +"~std.string.toString(bonus));
        
        free();
    }
}




// enemy missile
class CBullet : public IActor
{
public:
    static CSolidModel mdl;
    
    mixin TInstanceFunction!(CBullet);
    
    float   abilityHoming;
    bool    flagPass;
    
    CBullet initialize(float x, float y, float z, float v, float ability)
    {
        set(&pos, x, y, z);
        set(&vel, 0, 0, v);
        abilityHoming = ability;
        flagPass = false;
        return this;
    }
    
    void run()
    {
        vector3 require_vel;
        
        // control
        if (pos.z > -100) {
            copy(&require_vel, &vel);
            // hit test
            if (!flagPass) {
                flagPass = true;
                if (testHit()) {
                    // Hit !!
                    flyer.destroyed();
                }
            }
        } else {
            // frame count to achieve player z position.
            float frm_inv = (vel.z + flyer.global_vel.z) / pos.z;
            // require_vel; required velocity to aim player
            set(&require_vel, (pos.x-flyer.pos.x)*frm_inv-flyer.global_vel.x, (pos.y-flyer.pos.y)*frm_inv-flyer.global_vel.y, 0);
        }
        
        // calc aiming speed (accelaration)
        scale(sub(&acc.v2, &require_vel.v2, &vel.v2), 0.2);
        if (acc.x >  abilityHoming) acc.x =  abilityHoming;
        if (acc.x < -abilityHoming) acc.x = -abilityHoming;
        if (acc.y >  abilityHoming) acc.y =  abilityHoming;
        if (acc.y < -abilityHoming) acc.y = -abilityHoming;

        // physics
        add(&pos, add(&vel, &acc));
        add(&pos, &flyer.global_vel);
        sortZ();
        
        // posture(bank, head, pitch)
        vector3 dir;
        normalize(set(&dir, vel.x, vel.y, vel.z+flyer.vel.z));
        set(&posture, 0, radian2degree(asin(dir.x)), radian2degree(asin(-dir.y)));

        // smoke
        float dst = 160;
        CSmoke().initialize(pos.x-dir.x*dst, pos.y-dir.y*dst, pos.z-dir.z*dst, BULLET_SMOKE_LENGTH, 160, gradSmoke)
                .insert_before(this);
        if (pos.z > -60*BULLET_SMOKE_LENGTH) {
            dst += vel.z*0.5;
            CSmoke().initialize(pos.x-dir.x*dst, pos.y-dir.y*dst, pos.z-dir.z*dst, BULLET_SMOKE_LENGTH, 160, gradSmoke)
                    .insert_before(this);
        }
        
        // disapeared
        if (pos.z > 500) free();
    }
    
    void draw()
    {
        setModelAttribution();
        glDisable(GL_FOG);
        glPushMatrix();
            glTranslatef(pos.x, pos.y, pos.z);
            glRotatef(posture.z, 1, 0, 0);  // pitch
            glRotatef(posture.y, 0, 1, 0);  // head
            mdl.callList();
        glPopMatrix();
    }

    // call from run()
    bool testHit()
    {
        return (flyer.isAlive() && calcDistance2(&flyer.pos.v2, &pos.v2) < 2500);
    }
}




// explosion
class CExplosion : public IActor
{
public:
    static CSprite spr;
    
    mixin TInstanceFunction!(CExplosion);
    
    float   size;
    float   rotation;
    int     timer;
    
    CExplosion initialize(float x, float y, float z, float size_)
    {
        set(&pos, x, y, z);
        size = size_;
        rotation = rand.getsf(180);
        timer = 0;
        return this;
    }
    
    void run()
    {
        add(&pos, &flyer.global_vel);
        if (++timer > 48) free();
    }
    
    void draw()
    {
        setSpriteAttribution();
        glDisable(GL_FOG);
        setAlphaBlend();
        glColor4f(1,1,1,1);
        glPushMatrix();
            glTranslatef(pos.x, pos.y, pos.z);
            int u = timer & 7;
            int v = timer>>3;
            spr.setUV(u*0.125, v*0.125, 0.125, 0.125);
            spr.set(size, size);
            spr.rotate(rotation);
            spr.draw();
        glPopMatrix();
    }
}




// explosion particle
class CExplosionParticle : public IActor
{
public:
    mixin TInstanceFunction!(CExplosionParticle);
    int   timer;
    float size;
    
    CExplosionParticle initialize(float x, float y, float z, float size_)
    {
        set(&pos, x, y, z);
        timer = 0;
        size = size_;
        return this;
    }
    
    void run()
    {
        add(&pos, &flyer.global_vel);
        for(int i=0; i<3; i++) {
            CSmoke().initialize(pos.x+rand.getsf(size), pos.y+rand.getsf(size), pos.z, rand.geti(60)+90, size*1.5, gradDestroy).insert_after(this);
        }
        if (++timer == 10) free();
    }
    
    void draw()
    {
    }
}




// smoke
class CSmoke : public IActor
{
public:
    static CSprite spr;
    
    mixin TInstanceFunction!(CSmoke);
    
    float       life;
    float       aging;
    float       rotation;
    float       size;
    CGradation  grad;
    
    CSmoke initialize(float x, float y, float z, float life_, float size_, CGradation grad_)
    {
        set(&pos, x, y, z);
        set(&vel, rand.getsf(0.3), rand.getsf(0.3), 0);
        aging = 1/life_;
        life = 1;
        scale(&acc, &vel, -aging);
        rotation = rand.getsf(180);
        size = size_;
        grad = grad_;
        return this;
    }
    
    void run()
    {
        // physics
        add(add(&pos, &vel), &acc, 0.5);
        add(&vel, &acc);
        add(&pos, &flyer.global_vel);
        
        // disapeared
        life -= aging;
        if (life < 0 || pos.z > 500) free();
    }
    
    void draw()
    {
        setSpriteAttribution();
        glDisable(GL_FOG);
        setAlphaBlend();
        glColor4fv(&grad(1-life).r);
        glPushMatrix();
            glTranslatef(pos.x, pos.y, pos.z);
            spr.set(size, size);
            spr.rotate(rotation);
            spr.draw();
        glPopMatrix();
    }
}




// sky
class CSky : public IActor
{
public:
    CTexture     bg_tex;
    CCircleArray bg;
    CSolidModel  mdl;
    
    this()
    {
        bg_tex = (new CTexture("resource\\bgsky.bmp",3));
        bg_tex.register();
        bg = new CCircleArray(16, true, false, false);
        bg.setCylinder(2000, -500, 2000).setUV(0.2,0,2,1).setAxis(0,1,0);
        mdl = new CSolidModel("resource\\sky.mqo");
        mdl.scale(5, 5, 5);
    }
    
    void run()
    {
        // scrolling
        add(&pos, &flyer.global_vel);
        if (pos.x > 500) pos.x -= 1000;
        else if (pos.x < -500) pos.x += 1000;
        if (pos.z > 1000) pos.z -= 1000;
        else if (pos.z < 0) pos.z += 1000;
    }

    void draw()
    {
        glColor4f(1,1,1,1);
        setModelAttribution();
        glDisable(GL_LIGHTING);
        glDisable(GL_FOG);
        bg_tex.bind();
        bg.draw();
        bg_tex.unbind();

        glEnable(GL_FOG);
        glPushMatrix();
            glTranslatef(pos.x-1000, -250, pos.z);
            for (int i=0; i<3; i++) {
                mdl.draw();
                glTranslatef(1000, 0, 0);
                mdl.draw();
                glTranslatef(1000, 0, 0);
                mdl.draw();
                glTranslatef(-2000, 0, -1000);
            }
        glPopMatrix();
    }
}




