//------------------------------------------------------------------------------------------
// OpenGL Light Class
//------------------------------------------------------------------------------------------
private import opengl;
private import openglu;
private import util.vector;




class CLight
{
// member values
//--------------------------------------------------
public:
    vector3     position  = {x:0, y:-1, z:0};   // position
    vector3     direction = {x:0, y:1,  z:0};   // spot light dierction
    bool        isSpotLight = false;            // flag for spotlight
    GLenum      numLight = 0;                   // light number
    



// constructor
//--------------------------------------------------
public:
    // constructor set nothing
    this()
    {
        genLight();
    }
    

    // constructor for point light
    this(float x,  float y,  float z)
    {
        create(x, y, z);
    }

    
    // constructor for spot light
    this(float x,  float y,  float z, float dx, float dy, float dz)
    {
        create(x, y, z, dx, dy, dz);
    }

    
    // destructor
    ~this()
    {
        destroy();
    }



// initialize
//--------------------------------------------------
public:
    // create light
    CLight create(float x, float y, float z)
    {
        if (genLight()) {
            setPosition(x,y,z);
            on();
        }
        return this;
    }


    // create spot light
    CLight create(float x, float y, float z, float dx, float dy, float dz)
    {
        if (genLight()) {
            setPosition(x, y, z);
            setDirection(dx, dy, dz);
            on();
        }
        return this;
    }


    // destroy light
    void destroy()
    {
        if (numLight != 0) { off(); }
    }


    // generate light
    bool genLight()
    {
        for (int i=GL_LIGHT0; i<=GL_LIGHT7; i++) {
            if (!glIsEnabled(i)) {
                numLight = cast(GLenum)(i);
                return true;
            }
        }
        numLight = 0;
        return false;
    }




//  setting
//--------------------------------------------------
public:
    // set position
    CLight setPosition(float x, float y, float z)
    {
        set(&position, x, y, z);
        glLightfv(numLight, GL_POSITION, &position.x);
        return this;
    }


    // set spot light direction
    CLight setDirection(float dir_x, float dir_y, float dir_z)
    {
        isSpotLight = true;
        set(&direction, dir_x, dir_y, dir_z);
        glLightfv(numLight, GL_SPOT_DIRECTION, &direction.x);
        return this;
    }


    // set color
    CLight setColor(uint color, float amb, float diff, float spec)
    {
        vector4 v, l;
        setUint(&v, color);
        scale(&l, &v, amb);
        glLightfv(numLight, GL_AMBIENT, &l.x);
        scale(&l, &v, diff);
        glLightfv(numLight, GL_DIFFUSE, &l.x);
        scale(&l, &v, spec);
        glLightfv(numLight, GL_SPECULAR, &l.x);
        return this;
    }


    // set spot light attribution
    CLight setSpotLight(float angle, float exp, float t0, float t1, float t2)
    {
        isSpotLight = true;
        glLightf(numLight, GL_SPOT_CUTOFF,    angle);
        glLightf(numLight, GL_SPOT_EXPONENT,  exp);
        glLightf(numLight, GL_CONSTANT_ATTENUATION,  t0);
        glLightf(numLight, GL_LINEAR_ATTENUATION,    t1);
        glLightf(numLight, GL_QUADRATIC_ATTENUATION, t2);
        return this;
    }


    // update position and direction
    void update()
    {
        glLightfv(numLight, GL_POSITION, &position.x);
        glLightfv(numLight, GL_SPOT_DIRECTION, &direction.x);
    }




// operation
//--------------------------------------------------
    void on()
    {
        glEnable(numLight);
    }


    void off()
    {
        glDisable(numLight);
    }
};
